#include "bvp.hh"

/*


 */

/*

  Solve BVP
  
     y''(x) + p(x) y'(x) + q(x) y(x) = r(x)
     y(a) = Ya, y(b) = Yb

  =================

   y[k-1]-2*y[k]+y[k+1]            y[k+1]-y[k-1]
   --------------------  + p(x[k]) ------------- + q(x[k]) * y[k] = r(x[k]) 
          h*h                           2*h

   y[0] = Ya
   y[N] = Yb

  =================

   y[k-1] * ( 1 - h*p(x[k])/2 ) +
   y[k]   * ( -2 + (h*h)*q(x[k]) ) +
   y[k+1] * ( 1 + h*p(x[k])/2 ) = (h*h)*r(x[k])

  =================

   alpha[k] * y[k-1] + beta[k] * y[k] + gamma[k] * y[k+1] = omega[k]

   alpha[k] = 1 - h*p(x[k])/2 ;
   beta[k]  = -2 + (h*h)*q(x[k]) ;
   gamma[k] = 1 + h*p(x[k])/2 ;
   omega[k] = (h*h)*r(x[k]) ;

  =================

*/

using namespace std ;

void
BVP::solve( value_type a, value_type Ya,
            value_type b, value_type Yb,
            int        N ) {
  vector<value_type> alpha(N+1), beta(N+1), gamma(N+1), omega(N+1) ;
  value_type h = (b-a)/N ;
  // equation of the first BC
  alpha[0] = 0 ;
  beta[0]  = 1 ;
  gamma[0] = 0 ;
  omega[0] = Ya ;
  // equation of the second BC
  alpha[N] = 0 ;
  beta[N]  = 1 ;
  gamma[N] = 0 ;
  omega[N] = Yb ;
  
  // fill the lienear system
  value_type h2 = h*h ;
  x.resize(N+1) ;
  y.resize(N+1) ;
  x[0] = a ;
  x[N] = b ;

  for ( int k = 1 ; k < N ; ++k ) {
    value_type xk = a+h*k  ;
    value_type bf = h*p(xk)/2 ;
    x[k]     = xk ;
    alpha[k] = 1 - bf ;
    beta[k]  = -2 + h2*q(xk)  ;  
    gamma[k] = 1 + bf ;
    omega[k] = h2*r(xk) ;
  }

  // solve linear system and compute y
  // eliminate lower diagonal
  for ( int k = 1 ; k <= N ; ++k ) {
    value_type bf = alpha[k]/beta[k-1] ;
    beta[k]  -= bf*gamma[k-1] ;
    omega[k] -= bf*omega[k-1] ;
  }

  // solve bidiagonal system
  y[N] = omega[N] / beta[N] ;
  for ( int k = N-1 ; k >= 0 ; --k ) y[k] = (omega[k] - gamma[k]*y[k+1]) / beta[k] ;

}

